from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import keras as K
from keras import regularizers
from keras.layers import BatchNormalization, Activation
from keras.layers import Conv2D, Conv3D
from keras.layers.convolutional_recurrent import ConvLSTM2D
from keras import layers
from keras import models
from keras.layers.core import Lambda
from keras.layers import  TimeDistributed, Reshape



reg_weight = 1.e-4

"""
Using RESNET and stacked layer style architechtures to define NEAT architecture
"""



class Concat(layers.Layer):

     def __init__(self, axis = -1, name = 'Concat', **kwargs):

          self.axis = axis
          super(Concat, self).__init__(name = name)

     def get_config(self):
        config = super().get_config()
        config.update({
            "axis": self.axis,
            "name": self.name,
        })
        return config
 
     def call(self, x):

        y = Lambda(lambda x:layers.concatenate([x[0], x[1]], self.axis))(x)

        return y 






def TDresnet_layer(inputs,
                 num_filters=64,
                 kernel_size= 3,
                 strides=1,
                 activation='relu',
                 batch_normalization=True,
                 conv_first=True):
    """2D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    
    conv = TimeDistributed(Conv2D(num_filters,
                  kernel_size=kernel_size,
                  strides=strides,
                  padding='same',
                  kernel_initializer='he_normal',
                  kernel_regularizer=regularizers.l2(1e-4)))

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = TimeDistributed(BatchNormalization())(x)
        if activation is not None:
            x = TimeDistributed(Activation(activation))(x)
    else:
        if batch_normalization:
            x = TimeDistributed(BatchNormalization())(x)
        if activation is not None:
            x = TimeDistributed(Activation(activation))(x)
        x = conv(x)
    return x








def OSNET(input_shape, categories,unit, box_vector, nboxes = 1, stage_number = 3, last_conv_factor = 4, depth = 38, start_kernel = 3, mid_kernel = 3, lstm_kernel = 3, startfilter = 48,  input_weights = None,last_activation = 'softmax'):
    """ResNet Version 2 Model builder [b]
    depth of 29 == max pooling of 28 for image patch of 55
    depth of 56 == max pooling of 14 for image patch of 55
    """
    img_input = layers.Input(shape = (input_shape[0], None, None, input_shape[3]))
    if (depth - 2) % 9 != 0:
        raise ValueError('depth should be 9n+2 (eg 56 or 110 in [b])')
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = ThreeDresnet_layer(inputs=img_input,
                     num_filters=num_filters_in,
                     kernel_size = start_kernel,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2   # downsample

            # bottleneck residual unit
  
            x = ThreeDresnet_layer(inputs=x,
                             num_filters=num_filters_in,
                               kernel_size= mid_kernel,
                             conv_first=False,
                             strides=strides,
                             activation=activation,
                             batch_normalization=batch_normalization)


              
        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    num_filters_in_TD = startfilter
    num_res_blocks = int((depth - 2) / 9)
    
    z = TDresnet_layer(inputs=img_input,
                     num_filters=num_filters_in_TD,
                     kernel_size = start_kernel,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in_TD * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in_TD * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2   # downsample

            # bottleneck residual unit
      
            z = TDresnet_layer(inputs=z,
                             num_filters=num_filters_in_TD,
                               kernel_size= mid_kernel,
                             conv_first=False,
                             strides=strides,
                             activation=activation,
                             batch_normalization=batch_normalization)


              
        num_filters_in_TD = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    z = BatchNormalization()(z)
    z = Activation('relu')(z)
    
    
    
    branchAdd = layers.add([z, x])
    

    
    x = ConvLSTM2D(filters = unit, kernel_size = (lstm_kernel, lstm_kernel),  activation='relu', data_format = 'channels_last', return_sequences = False, padding = "same", name = "newlstmdeep")(branchAdd)


    x = (Conv2D(categories + nboxes * box_vector, kernel_size= mid_kernel,kernel_regularizer=regularizers.l2(reg_weight), padding = 'same'))(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    input_cat = Lambda(lambda x:x[:,:,:,0:categories])(x)
    input_box = Lambda(lambda x:x[:,:,:,categories:])(x)

        
    output_cat = (Conv2D(categories, (round(input_shape[1]/last_conv_factor),round(input_shape[2]/last_conv_factor)),activation= last_activation ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid', name = 'yolo'))(input_cat)
    output_box = (Conv2D(nboxes * (box_vector), (round(input_shape[1]/last_conv_factor),round(input_shape[2]/last_conv_factor)),activation= 'sigmoid' ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid', name = 'secyolo'))(input_box)




    block = Concat(-1)
    outputs = block([output_cat,output_box]) 
    
    
    
    inputs = img_input
   
    # Create model.
    model = models.Model(inputs, outputs)


    if input_weights is not None:

       model.load_weights(input_weights, by_name =True)
    
    return model


def ORNET(input_shape, categories,unit, box_vector,nboxes = 1, stage_number = 3, last_conv_factor = 4, depth = 38, start_kernel = 3, mid_kernel = 3, lstm_kernel = 3, startfilter = 32,  input_weights = None, last_activation = 'softmax'):
    """ResNet Version 2 Model builder [b]
    depth of 29 == max pooling of 28 for image patch of 55
    depth of 56 == max pooling of 14 for image patch of 55
    """
    img_input = layers.Input(shape = (input_shape[0], None, None, input_shape[3]))
    if (depth - 2) % 9 != 0:
        raise ValueError('depth should be 9n+2 (eg 56 or 110 in [b])')
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = ThreeDresnet_layer(inputs=img_input,
                     num_filters=num_filters_in,
                     kernel_size = start_kernel,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2   # downsample

            # bottleneck residual unit
            y = ThreeDresnet_layer(inputs=x,
                             num_filters=num_filters_in,
                             kernel_size=1,
                             strides=strides,
                             activation=activation,
                             batch_normalization=batch_normalization,
                             conv_first=False)
            y = ThreeDresnet_layer(inputs=y,
                             num_filters=num_filters_in,
                               kernel_size= mid_kernel,
                             conv_first=False)
            y = ThreeDresnet_layer(inputs=y,
                             num_filters=num_filters_out,
                             kernel_size=1,
                             conv_first=False)
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = ThreeDresnet_layer(inputs=x,
                                 num_filters=num_filters_out,
                                 kernel_size=1,
                                 strides=strides,
                                 activation=None,
                                 batch_normalization=False)
              
            x = K.layers.add([x, y])
        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    num_filters_in_TD = startfilter
    num_res_blocks = int((depth - 2) / 9)
    
    z = TDresnet_layer(inputs=img_input,
                     num_filters=num_filters_in_TD,
                     kernel_size = start_kernel,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in_TD * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in_TD * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2   # downsample

            # bottleneck residual unit
            yz = TDresnet_layer(inputs=z,
                             num_filters=num_filters_in_TD,
                             kernel_size=1,
                             strides=strides,
                             activation=activation,
                             batch_normalization=batch_normalization,
                             conv_first=False)
            yz = TDresnet_layer(inputs=yz,
                             num_filters=num_filters_in_TD,
                               kernel_size= mid_kernel,
                             conv_first=False)
            yz = TDresnet_layer(inputs=yz,
                             num_filters=num_filters_out,
                             kernel_size=1,
                             conv_first=False)
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                z = TDresnet_layer(inputs=z,
                                 num_filters=num_filters_out,
                                 kernel_size=1,
                                 strides=strides,
                                 activation=None,
                                 batch_normalization=False)
              
            z = K.layers.add([z, yz])
        num_filters_in_TD = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    z = BatchNormalization()(z)
    z = Activation('relu')(z)
    
    
    
    branchAdd = layers.add([z, x])
    

    
    x = ConvLSTM2D(filters = unit, kernel_size = (lstm_kernel, lstm_kernel),  activation='relu', data_format = 'channels_last', return_sequences = False, padding = "same", name = "newlstmdeep")(branchAdd)


    x = (Conv2D(categories + nboxes * box_vector, kernel_size= mid_kernel,kernel_regularizer=regularizers.l2(reg_weight), padding = 'same'))(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    input_cat = Lambda(lambda x:x[:,:,:,0:categories])(x)
    input_box = Lambda(lambda x:x[:,:,:,categories:])(x)

        
    output_cat = (Conv2D(categories, (round(input_shape[1]/last_conv_factor),round(input_shape[2]/last_conv_factor)),activation= last_activation,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid', name = 'yolo'))(input_cat)
    output_box = (Conv2D(nboxes*(box_vector), (round(input_shape[1]/last_conv_factor),round(input_shape[2]/last_conv_factor)),activation= 'sigmoid' ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid', name = 'secyolo'))(input_box)




    block = Concat(-1)
    outputs = block([output_cat,output_box]) 
    
  
    inputs = img_input
   
    # Create model.
    model = models.Model(inputs, outputs)


    if input_weights is not None:

       model.load_weights(input_weights, by_name =True)
    
    return model









    
def ThreeDresnet_layer(inputs,
                 num_filters=64,
                 kernel_size= 3,
                 strides=1,
                 activation='relu',
                 batch_normalization=True,
                 conv_first=True):
    """2D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv3D(num_filters,
                  kernel_size=kernel_size,
                  strides=(1,strides,strides),
                  padding='same',
                  kernel_initializer='he_normal',
                  kernel_regularizer=regularizers.l2(1e-4))

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = (BatchNormalization())(x)
        if activation is not None:
            x = (Activation(activation))(x)
    else:
        if batch_normalization:
            x = (BatchNormalization())(x)
        if activation is not None:
            x = (Activation(activation))(x)
        x = conv(x)
    return x
    





def resnet_v2(input_shape, categories, box_vector,nboxes = 1, stage_number = 3, last_conv_factor = 4,  depth = 38,  start_kernel = 3, mid_kernel = 3, startfilter = 48,  input_weights = None, last_activation = 'softmax'):
    """ResNet Version 2 Model builder [b]
    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """
    img_input = layers.Input(shape = (None, None, input_shape[2]))
    if (depth - 2) % 9 != 0:
        raise ValueError('depth should be 9n+2 (eg 56 or 110 in [b])')
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_layer(inputs=img_input,
                     num_filters=num_filters_in,
                     kernel_size = start_kernel,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_layer(inputs=x,
                             num_filters=num_filters_in,
                             kernel_size=1,
                             strides=strides,
                             activation=activation,
                             batch_normalization=batch_normalization,
                             conv_first=False)
            y = resnet_layer(inputs=y,
                             num_filters=num_filters_in,
                               kernel_size= mid_kernel,
                             conv_first=False)
            y = resnet_layer(inputs=y,
                             num_filters=num_filters_out,
                             kernel_size=1,
                             conv_first=False)
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_layer(inputs=x,
                                 num_filters=num_filters_out,
                                 kernel_size=1,
                                 strides=strides,
                                 activation=None,
                                 batch_normalization=False)
              
            x = K.layers.add([x, y])
        
        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = (Conv2D(categories + nboxes * box_vector, kernel_size= mid_kernel,kernel_regularizer=regularizers.l2(reg_weight), padding = 'same'))(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    

    
    input_cat = Lambda(lambda x:x[:,:,:,0:categories])(x)
    input_box = Lambda(lambda x:x[:,:,:,categories:])(x)
    
      
    

    output_cat = (Conv2D(categories, (round(input_shape[0]/last_conv_factor),round(input_shape[1]/last_conv_factor)),activation= last_activation ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid'))(input_cat)
    output_box = (Conv2D(nboxes * (box_vector), (round(input_shape[0]/last_conv_factor),round(input_shape[1]/last_conv_factor)),activation= 'sigmoid' ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid'))(input_box)
    

    block = Concat(-1)
    outputs = block([output_cat,output_box])
    
    inputs = img_input
   
     
    # Create model.
    model = models.Model(inputs, outputs)
    
    
    if input_weights is not None:

        model.load_weights(input_weights, by_name =True)
        
    return model
  
def seqnet_v2(input_shape, categories, box_vector,nboxes = 1,stage_number = 3, last_conv_factor = 4, depth = 38, start_kernel = 3, mid_kernel = 3, startfilter = 48,  input_weights = None, last_activation = 'softmax'):
    """ResNet Version 2 Model builder [b]
    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """
    img_input = layers.Input(shape = (None, None, input_shape[2]))
    if (depth - 2) % 9 != 0:
        raise ValueError('depth should be 9n+2 (eg 56 or 110 in [b])')
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_layer(inputs=img_input,
                     num_filters=num_filters_in,
                     kernel_size = start_kernel,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit

            x = resnet_layer(inputs=x,
                             num_filters=num_filters_in,
                               kernel_size= mid_kernel,
                             conv_first=False,
                             strides = strides,
                             activation=activation,
                             batch_normalization=batch_normalization)
           
 
              
        
        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = (Conv2D(categories + nboxes * box_vector, kernel_size= mid_kernel,kernel_regularizer=regularizers.l2(reg_weight), padding = 'same'))(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    

    
    input_cat = Lambda(lambda x:x[:,:,:,0:categories])(x)
    input_box = Lambda(lambda x:x[:,:,:,categories:])(x)
    
      
    

    output_cat = (Conv2D(categories, (round(input_shape[0]/last_conv_factor),round(input_shape[1]/last_conv_factor)),activation= last_activation ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid'))(input_cat)
    output_box = (Conv2D(nboxes*(box_vector), (round(input_shape[0]/last_conv_factor),round(input_shape[1]/last_conv_factor)),activation= 'sigmoid' ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid'))(input_box)
    

    block = Concat(-1)
    outputs = block([output_cat,output_box])

    
    inputs = img_input
   
     
    # Create model.
    model = models.Model(inputs, outputs)
    
    
    if input_weights is not None:

        model.load_weights(input_weights, by_name =True)
        
    return model
      

def resnet_3D_v2(input_shape, categories,box_vector, stage_number = 3, last_conv_factor = 4,  depth = 38,  start_kernel = 3, mid_kernel = 3, startfilter = 48,  input_weights = None, last_activation = 'softmax'):
    """ResNet Version 2 Model builder [b]
    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """
    img_input = layers.Input(shape = (input_shape[0], None, None, input_shape[3]))
    if (depth - 2) % 9 != 0:
        raise ValueError('depth should be 9n+2 (eg 56 or 110 in [b])')
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_3D_layer(inputs=img_input,
                     num_filters=num_filters_in,
                     kernel_size = start_kernel,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_3D_layer(inputs=x,
                             num_filters=num_filters_in,
                             kernel_size=1,
                             strides=strides,
                             activation=activation,
                             batch_normalization=batch_normalization,
                             conv_first=False)
            y = resnet_3D_layer(inputs=y,
                             num_filters=num_filters_in,
                               kernel_size= mid_kernel,
                             conv_first=False)
            y = resnet_3D_layer(inputs=y,
                             num_filters=num_filters_out,
                             kernel_size=1,
                             conv_first=False)
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_3D_layer(inputs=x,
                                 num_filters=num_filters_out,
                                 kernel_size=1,
                                 strides=strides,
                                 activation=None,
                                 batch_normalization=False)
              
            x = K.layers.add([x, y])
        
        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = (Conv3D(categories+  box_vector, kernel_size= mid_kernel,kernel_regularizer=regularizers.l2(reg_weight), padding = 'same'))(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    
    input_cat = Lambda(lambda x:x[:,:,:,:,0:categories])(x)
    input_box = Lambda(lambda x:x[:,:,:,:,categories:])(x)
    
      
    

    output_cat = (Conv3D(categories, (round(input_shape[0]),round(input_shape[1]/last_conv_factor), round(input_shape[2]/last_conv_factor)),activation= last_activation ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid'))(input_cat)
    output_box = (Conv3D((box_vector), (round(input_shape[0]),round(input_shape[1]/last_conv_factor), round(input_shape[2]/last_conv_factor)),activation= 'sigmoid' ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid'))(input_box)
    

    block = Concat(-1)
    outputs = block([output_cat,output_box])


    
    inputs = img_input
   
    # Create model.
    model = models.Model(inputs, outputs)
    
    
    if input_weights is not None:

        model.load_weights(input_weights, by_name =True)
        
    return model
  
def seqnet_3D_v2(input_shape, categories, box_vector, stage_number = 3, last_conv_factor = 4, depth = 38, start_kernel = 3, mid_kernel = 3, startfilter = 48,  input_weights = None, last_activation = 'softmax'):
    """ResNet Version 2 Model builder [b]
    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """
    img_input = layers.Input(shape = (input_shape[0], None, None, input_shape[3]))
    if (depth - 2) % 9 != 0:
        raise ValueError('depth should be 9n+2 (eg 56 or 110 in [b])')
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_3D_layer(inputs=img_input,
                     num_filters=num_filters_in,
                     kernel_size = start_kernel,
                     conv_first=True)

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = 'relu'
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit

            x = resnet_3D_layer(inputs=x,
                             num_filters=num_filters_in,
                               kernel_size= mid_kernel,
                             conv_first=False,
                             strides = strides,
                             activation=activation,
                             batch_normalization=batch_normalization)
           
 
              
        
        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = (Conv3D(categories+ box_vector, kernel_size= mid_kernel,kernel_regularizer=regularizers.l2(reg_weight), padding = 'same'))(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    

    

    input_cat = Lambda(lambda x:x[:,:,:,:,0:categories])(x)
    input_box = Lambda(lambda x:x[:,:,:,:,categories:])(x)
    
      
    

    output_cat = (Conv3D(categories, (round(input_shape[0]),round(input_shape[1]/last_conv_factor), round(input_shape[2]/last_conv_factor)),activation= last_activation ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid'))(input_cat)
    output_box = (Conv3D((box_vector), (round(input_shape[0]),round(input_shape[1]/last_conv_factor), round(input_shape[2]/last_conv_factor)),activation= 'sigmoid' ,kernel_regularizer=regularizers.l2(reg_weight), padding = 'valid'))(input_box)
    

    block = Concat(-1)
    outputs = block([output_cat,output_box])

    
    inputs = img_input
   
     
    # Create model.
    model = models.Model(inputs, outputs)
    
    
    if input_weights is not None:

        model.load_weights(input_weights, by_name =True)
        
    return model
      


    
   
def resnet_layer(inputs,
                 num_filters=64,
                 kernel_size=3,
                 strides=1,
                 activation='relu',
                 batch_normalization=True,
                 conv_first=True):
    """2D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv2D(num_filters,
                  kernel_size=kernel_size,
                  strides=strides,
                  padding='same',
                  kernel_initializer='he_normal',
                  kernel_regularizer=regularizers.l2(1e-4))

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x
    
def resnet_3D_layer(inputs,
                 num_filters=64,
                 kernel_size=3,
                 strides=1,
                 activation='relu',
                 batch_normalization=True,
                 conv_first=True):
    """3D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv3D(num_filters,
                  kernel_size=kernel_size,
                  strides=(1,strides,strides),
                  padding='same',
                  kernel_initializer='he_normal',
                  kernel_regularizer=regularizers.l2(1e-4))

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x
    





from qtpy.QtWidgets import QWidget, QHBoxLayout

from qtpy.QtWidgets import QListWidget

# NOT IMPLEMENTED YET
class KeyFramesListWidget(QListWidget):
    """List of Key Frames."""

    def __init__(self, animation, parent=None):
        super().__init__(parent=parent)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 30 16:05:34 2021
@author: vkapoor
"""

import tensorflow as tf
import numpy as np
from keras import backend as K

lambdaobject = 1
lambdanoobject = 1
lambdacoord = 1
lambdaclass = 1
lambdaangle = 1

def get_event_grid(grid_h, grid_w, grid_t, boxes):
    
    event_grid = np.array([ [[float(y),float(x),float(t)]]*boxes  for y in range(grid_h) for x in range(grid_w) for t in range(grid_t)])
    
    return event_grid


def extract_ground_event_truth(y_true, categories, grid_h, grid_w, grid_t, nboxes, box_vector,yolo_v0,yolo_v1,yolo_v2):
    
    true_box_class = y_true[...,0:categories]
    
    true_nboxes = K.reshape(y_true[...,categories:], (-1, grid_h * grid_w * grid_t, nboxes, box_vector))
    
    true_box_xyt = true_nboxes[...,0:3]
    true_box_t = K.ones(true_nboxes[...,3])

    true_box_wh =  tf.keras.layers.Concatenate(axis=1)([true_nboxes[...,3:5], true_box_t]) 
    
    if yolo_v0:
        true_box_conf = 1
        true_box_angle = 2
        
    
    else:
            if yolo_v1 or yolo_v2:    
               true_box_conf = true_nboxes[...,5]
               true_box_angle = 2
            if yolo_v2:
                true_box_angle = true_nboxes[...,6]
                
    return true_box_class, true_box_xyt, true_box_wh, true_box_conf, true_box_angle 
               
    
def extract_ground_event_pred(y_pred, categories, grid_h, grid_w, grid_t, event_grid, nboxes, box_vector, yolo_v0,yolo_v1,yolo_v2):
    
    pred_box_class = y_pred[...,0:categories]
    
    pred_nboxes = K.reshape(y_pred[...,categories:], (-1, grid_h * grid_w * grid_t, nboxes, box_vector))
    
    pred_box_xyt = pred_nboxes[...,0:3] + event_grid
    pred_box_t = K.ones(pred_nboxes[...,3])
    pred_box_wh = tf.keras.layers.Concatenate(axis=1)([pred_nboxes[...,3:5], pred_box_t])
        
    if yolo_v0:
        
        pred_box_conf = 1
        pred_box_angle = 2
        
    
    else:
            if yolo_v1 or yolo_v2:    
               pred_box_conf = pred_nboxes[...,5]
               pred_box_angle = 2
            if yolo_v2:
                pred_box_angle = pred_nboxes[...,6]
                
    return pred_box_class, pred_box_xyt, pred_box_wh, pred_box_conf, pred_box_angle 
            
    
def get_cell_grid(grid_h, grid_w, boxes):
    
    cell_grid = np.array([ [[float(x),float(y)]]*boxes   for y in range(grid_h) for x in range(grid_w)])
    
    return cell_grid

    
def extract_ground_cell_pred(y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector, yolo_v0):

        pred_box_class = y_pred[...,0:categories]
        
        pred_nboxes = K.reshape(y_pred[...,categories:], (-1, grid_h * grid_w, nboxes, box_vector))
    
        pred_box_xy = pred_nboxes[...,0:2] + K.variable(cell_grid)
    
        pred_box_wh = pred_nboxes[...,2:4]
        
        if yolo_v0:
            
                pred_box_conf = 1
         
        else: 
        
                pred_box_conf = pred_nboxes[...,4]
        
        return pred_box_class, pred_box_xy, pred_box_wh, pred_box_conf

def extract_ground_cell_truth(y_truth, categories, grid_h, grid_w, nboxes, box_vector, yolo_v0):

        true_box_class = y_truth[...,0:categories]
        
        true_nboxes = K.reshape(y_truth[...,categories:], (-1, grid_h * grid_w, nboxes, box_vector))
    
        true_box_xy = true_nboxes[...,0:2]
    
        true_box_wh = true_nboxes[...,2:4]
        
        if yolo_v0:
             
               true_box_conf = 1
         
        else: 
        
               true_box_conf = true_nboxes[...,4]
        
        return true_box_class, true_box_xy, true_box_wh, true_box_conf


def extract_ground_cell_pred_segfree(y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector):

        pred_box_class = y_pred[...,0:categories]
        
        pred_nboxes = K.reshape(y_pred[...,categories:], (-1, grid_h * grid_w, nboxes, box_vector))
    
        pred_box_xy = pred_nboxes[...,0:2] + cell_grid
    
        
        return pred_box_class, pred_box_xy

def extract_ground_cell_truth_foc(y_truth, categories):

        true_box_class = y_truth[...,0:categories]
        
      
    
        
        return true_box_class

def extract_ground_cell_pred_foc(y_pred, categories):

        pred_box_class = y_pred[...,0:categories]
       
    
        
        return pred_box_class

def extract_ground_cell_truth_segfree(y_truth, categories, grid_h, grid_w, nboxes, box_vector):

        true_box_class = y_truth[...,0:categories]
        
        true_nboxes = K.reshape(y_truth[...,categories:], (-1, grid_h * grid_w, nboxes, box_vector))
    
        true_box_xy = true_nboxes[...,0:2]
    
        
        return true_box_class, true_box_xy
def compute_conf_loss(pred_box_wh, true_box_wh, pred_box_xy,true_box_xy,true_box_conf,pred_box_conf):
    
# compute the intersection of all boxes at once (the IOU)
        intersect_wh = K.maximum(K.zeros_like(pred_box_wh), (pred_box_wh + true_box_wh)/2 - K.square(pred_box_xy- true_box_xy) )
        intersect_area = intersect_wh[...,0] * intersect_wh[...,1] * intersect_wh[...,2]
        true_area = true_box_wh[...,0] * true_box_wh[...,1] * true_box_wh[...,2]
        pred_area = pred_box_wh[...,0] * pred_box_wh[...,1] * pred_box_wh[...,2]
        union_area = pred_area + true_area - intersect_area
        iou = intersect_area / union_area
  
        loss_conf = K.sum(K.square(true_box_conf*iou - pred_box_conf), axis=-1)

        loss_conf = loss_conf * lambdaobject

        return loss_conf 

def compute_conf_loss_static(pred_box_wh, true_box_wh, pred_box_xy,true_box_xy,true_box_conf,pred_box_conf):
    
# compute the intersection of all boxes at once (the IOU)
        intersect_wh = K.maximum(K.zeros_like(pred_box_wh), (pred_box_wh + true_box_wh)/2 - K.square(pred_box_xy- true_box_xy) )
        intersect_area = intersect_wh[...,0] * intersect_wh[...,1] 
        true_area = true_box_wh[...,0] * true_box_wh[...,1] 
        pred_area = pred_box_wh[...,0] * pred_box_wh[...,1] 
        union_area = pred_area + true_area - intersect_area
        iou = intersect_area / union_area
  
        loss_conf = K.sum(K.square(true_box_conf*iou - pred_box_conf), axis=-1)

        loss_conf = loss_conf * lambdaobject

        return loss_conf        

def calc_loss_xywh(true_box_conf, true_box_xy, pred_box_xy, true_box_wh, pred_box_wh):

    
    loss_xy      = K.sum(K.sum(K.square(true_box_xy - pred_box_xy), axis = -1), axis = -1)
    loss_wh      = K.sum(K.sum(K.square(K.sqrt(true_box_wh) - K.sqrt(pred_box_wh)), axis=-1) * true_box_conf , axis=-1)
    loss_xywh = (loss_xy + loss_wh)
    loss_xywh = lambdacoord * loss_xywh
    return loss_xywh

def calc_loss_xy(true_box_xy, pred_box_xy):

    
    loss_xy      = K.sum(K.sum(K.square(true_box_xy - pred_box_xy), axis = -1), axis = -1)
    loss_xy = lambdacoord * loss_xy
    return loss_xy

def calc_loss_angle(true_box_conf, true_box_angle, pred_box_angle):

    
    loss_angle      = K.sum(K.sum(K.square(true_box_angle - pred_box_angle), axis = -1), axis = -1)
    loss_angle = lambdaangle * loss_angle
    return loss_angle
def calc_loss_class(true_box_class, pred_box_class, entropy):

    
        if entropy == 'binary':
            loss_class = K.mean(K.binary_crossentropy(true_box_class, pred_box_class), axis=-1)
        if entropy == 'notbinary':
            loss_class   = K.mean(K.categorical_crossentropy(true_box_class, pred_box_class), axis=-1)

        loss_class   = loss_class * lambdaclass 

        return loss_class



def dynamic_yolo_loss(categories, grid_h, grid_w, grid_t, nboxes, box_vector, entropy, yolo_v0, yolo_v1, yolo_v2):
    def loss(y_true, y_pred):    
        event_grid = get_event_grid(grid_h, grid_w, grid_t, nboxes)
        true_box_class, true_box_xyt, true_box_wh, true_box_conf, true_box_angle = extract_ground_event_truth(y_true, categories, grid_h, grid_w,grid_t, nboxes, box_vector, yolo_v0, yolo_v1, yolo_v2)
        pred_box_class, pred_box_xyt, pred_box_wh, pred_box_conf, pred_box_angle = extract_ground_event_pred(y_pred, categories, grid_h, grid_w,grid_t, event_grid, nboxes, box_vector, yolo_v0, yolo_v1, yolo_v2)

        loss_xywht = calc_loss_xywh(true_box_conf, true_box_xyt, pred_box_xyt, true_box_wh, pred_box_wh)

        loss_class   = calc_loss_class(true_box_class, pred_box_class, entropy)
        
        if yolo_v0:
                    combinedloss = (loss_xywht + loss_class)

        if yolo_v1:
                    loss_conf = compute_conf_loss(pred_box_wh, true_box_wh, pred_box_xyt,true_box_xyt,true_box_conf,pred_box_conf)
                    # Adding it all up   
                    combinedloss = (loss_xywht + loss_conf + loss_class)
        if yolo_v2:
                     
                    loss_conf = compute_conf_loss(pred_box_wh, true_box_wh, pred_box_xyt,true_box_xyt,true_box_conf,pred_box_conf) 
                    loss_angle = calc_loss_angle(true_box_conf, true_box_angle, pred_box_angle)                             
                    combinedloss = (loss_xywht + loss_conf + loss_class + loss_angle) 


        return combinedloss 
        
    return loss 


def static_yolo_loss(categories, grid_h, grid_w, nboxes, box_vector, entropy, yolo_v0):
    
    def loss(y_true, y_pred):    

        cell_grid = get_cell_grid(grid_h, grid_w, nboxes)
        true_box_class, true_box_xy, true_box_wh, true_box_conf = extract_ground_cell_truth(y_true, categories, grid_h, grid_w, nboxes, box_vector, yolo_v0)
        pred_box_class, pred_box_xy, pred_box_wh, pred_box_conf = extract_ground_cell_pred(y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector, yolo_v0)

        loss_xywh = calc_loss_xywh(true_box_conf, true_box_xy, pred_box_xy, true_box_wh, pred_box_wh)

        loss_class   = calc_loss_class(true_box_class, pred_box_class, entropy)
        
        if yolo_v0:
            
                    combinedloss = (loss_xywh + loss_class)

        else:
                    loss_conf = compute_conf_loss_static(pred_box_wh, true_box_wh, pred_box_xy,true_box_xy,true_box_conf,pred_box_conf)
                    # Adding it all up   
                    combinedloss = (loss_xywh + loss_conf + loss_class)



        return combinedloss 
        
    return loss    

def static_yolo_loss_segfree(categories, grid_h, grid_w, nboxes, box_vector, entropy, yolo_v0):
    
    def loss(y_true, y_pred):    

        cell_grid = get_cell_grid(grid_h, grid_w, nboxes)
        true_box_class, true_box_xy = extract_ground_cell_truth_segfree(y_true, categories, grid_h, grid_w, nboxes, box_vector, yolo_v0)
        pred_box_class, pred_box_xy = extract_ground_cell_pred_segfree(y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector, yolo_v0)

        loss_xy = calc_loss_xy(true_box_xy, pred_box_xy)

        loss_class   = calc_loss_class(true_box_class, pred_box_class, entropy)
        
                    # Adding it all up   
        combinedloss = (loss_xy + loss_class)



        return combinedloss 
        
    return loss  



from qtpy.QtWidgets import QWidget, QLabel, QVBoxLayout, QLineEdit, QPushButton

from ..animation import Animation
from ..easing import Easing
from .frame_widget import FrameWidget
import napari


class AnimationWidget(QWidget):
    """Widget for interatviely making animations using the napari viewer.

    Parameters
    ----------
    viewer : napari.Viewer
        napari viewer.

    Attributes
    ----------
    key_frames : list of dict
        List of viewer state dictionaries.
    frame : int
        Currently shown key frame.
    """

    def __init__(
        self,
        viewer: 'napari.viewer.Viewer',
        savedir: None,
        savename: None,
        start: 0,
        end: 1000,
        parent=None,
    ):
        super().__init__(parent=parent)

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._layout.addWidget(QLabel('Animation Wizard', parent=self))

        self.frameWidget = FrameWidget(parent=self)
        self._layout.addWidget(self.frameWidget)

        # self.captureButton = QPushButton('Capture Frame', parent=self)
        # self.captureButton.clicked.connect(self._capture_keyframe_callback)
        # self._layout.addWidget(self.captureButton)

        self._layout.addStretch(1)

        self.pathText = QLineEdit(parent=self)
        self.pathText.setText(savedir + savename + '.gif')
        self._layout.addWidget(self.pathText)

        self.saveButton = QPushButton('Save Animation', parent=self)
        self.saveButton.clicked.connect(self._save_callback)
        self._layout.addWidget(self.saveButton)

        self.frameWidget.endframeSpinBox.setRange(0, end)

        self.frameWidget.endframeSpinBox.setValue(end)
        self.frameWidget.startframeSpinBox.setValue(start)
        # Create animation
        
        self.animation = Animation(viewer, savedir)
        # establish key bindings
        self._add_callbacks()

    def _add_callbacks(self):
        """Bind keys"""

        self.animation.viewer.bind_key(
            "Alt-f", self._capture_keyframe_callback, overwrite=True
        )
        self.animation.viewer.bind_key(
            "Alt-r", self._replace_keyframe_callback, overwrite=True
        )
        self.animation.viewer.bind_key(
            "Alt-d", self._delete_keyframe_callback, overwrite=True
        )

        self.animation.viewer.bind_key("Alt-a", self._key_adv_frame, overwrite=True)
        self.animation.viewer.bind_key("Alt-b", self._key_back_frame, overwrite=True)

    def _release_callbacks(self):
        """Release keys"""

        self.animation.viewer.bind_key("Alt-f", None)
        self.animation.viewer.bind_key("Alt-r", None)
        self.animation.viewer.bind_key("Alt-d", None)

        self.animation.viewer.bind_key("Alt-a", None)
        self.animation.viewer.bind_key("Alt-b", None)

    def _get_interpolation_steps(self):
        return int(self.frameWidget.stepsSpinBox.value())

    def _get_easing_function(self):
        easing_name = str(self.frameWidget.easeComboBox.currentText())
        easing_func = Easing[easing_name.upper()].value
        return easing_func

    def _set_current_frame(self):
        return self.frameWidget.startframeSpinBox.setValue(self.animation.frame)

    def _capture_keyframe_callback(self, event=None):
        """Record current key-frame"""
        start = int(self.frameWidget.startframeSpinBox.value())
        end = int(self.frameWidget.endframeSpinBox.value())
        self.animation.capture_keyframe(start, end,
            steps=self._get_interpolation_steps(), ease=self._get_easing_function()
        )
        self._set_current_frame()

    def _replace_keyframe_callback(self, event=None):
        """Replace current key-frame with new view"""
        self.animation.capture_keyframe(
            steps=self._get_interpolation_steps(),
            ease=self._get_easing_function(),
            insert=False,
        )
        self._set_current_frame()

    def _delete_keyframe_callback(self, event=None):
        """Delete current key-frame"""

        self.animation.key_frames.pop(self.animation.frame)
        self.animation.frame = (self.animation.frame - 1) % len(
            self.animation.key_frames
        )
        self.animation.set_to_keyframe(self.animation.frame)
        self._set_current_frame()

    def _key_adv_frame(self, event=None):
        """Go forwards in key-frame list"""

        new_frame = (self.animation.frame + 1) % len(self.animation.key_frames)
        self.animation.set_to_keyframe(new_frame)
        self._set_current_frame()

    def _key_back_frame(self, event=None):
        """Go backwards in key-frame list"""

        new_frame = (self.animation.frame - 1) % len(self.animation.key_frames)
        self.animation.set_to_keyframe(new_frame)
        self._set_current_frame()

    def _save_callback(self, event=None):

        """Record current key-frame"""
        self._capture_keyframe_callback()

        path = self.pathText.text()
        self.animation.animate(path)

    def close(self):
        self._release_callbacks()
        super().close()




from .neat_goldstandard import NEATDynamic

class NEATSDynamic(NEATDynamic):
   

    def __init__(self, config, model_dir, model_name, catconfig=None, cordconfig=None):

                super().__init__(config = config, model_dir = model_dir, model_name = model_name, catconfig = catconfig, cordconfig = cordconfig)

    


    def predict_standard(self, imagename, markers, marker_tree, savedir, n_tiles=(1, 1), overlap_percent=0.8,
                event_threshold=0.5, iou_threshold=0.1, fidelity = 5, downsamplefactor = 1, watershed = None, maskimage = None, maskfilter = 10):

        self.predict(imagename,savedir,n_tiles = n_tiles, overlap_percent = overlap_percent, event_threshold = event_threshold, iou_threshold = iou_threshold, 
        fidelity = fidelity, downsamplefactor = downsamplefactor,  maskfilter = maskfilter, markers = markers, marker_tree = marker_tree, watershed = watershed, maskimage = maskimage,
         remove_markers = None )

def CreateVolume(patch, imaget, timepoint):
    starttime = timepoint
    endtime = timepoint + imaget
    smallimg = patch[starttime:endtime, :]

    return smallimg




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 15:25:10 2020

@author: aimachine
"""
import argparse
import numpy as np

class dynamic_config(argparse.Namespace):
    
    def __init__(self, npz_directory = None, npz_name = None, npz_val_name = None, key_categories = None, key_cord = None,  residual = True,stage_number = 3, last_conv_factor = 4, imagex = 128, imagey = 128, size_tminus = 3, size_tplus = 0, stride = 4,  nboxes = 1, depth = 29, start_kernel = 3, mid_kernel = 3, lstm_kernel = 3, startfilter = 48, lstm_hidden_unit = 16, epochs =100, learning_rate = 1.0E-4, batch_size = 10, model_name = 'NEATModel', yolo_v0 = True, yolo_v1 = False, yolo_v2 = False, multievent = False, show = True, **kwargs):
        
        
           self.npz_directory = npz_directory
           self.npz_name = npz_name
           self.npz_val_name = npz_val_name
           self.key_categories = key_categories
           self.key_cord = key_cord
           self.residual = residual
           self.yolo_v0 = yolo_v0
           self.yolo_v1 = yolo_v1
           self.yolo_v2 = yolo_v2
           self.nboxes = nboxes
           self.multievent = multievent
           self.categories = len(self.key_categories)
           self.box_vector = len(self.key_cord)
           self.depth = depth
           self.start_kernel = start_kernel
           self.mid_kernel = mid_kernel
           self.lstm_kernel = lstm_kernel
           self.startfilter = startfilter
           self.lstm_hidden_unit = lstm_hidden_unit
           self.epochs = epochs
           self.stride = last_conv_factor
           self.learning_rate = learning_rate
           self.batch_size = batch_size
           self.model_name = model_name
           self.show = show
           self.imagex = imagex
           self.imagey = imagey
           self.stage_number = stage_number
           self.last_conv_factor = last_conv_factor
           self.size_tminus = size_tminus
           self.size_tplus = size_tplus
           self.is_valid()
    

    def to_json(self):

         config = {
                 'npz_directory' : self.npz_directory,
                 'npz_name' : self.npz_name,
                 'npz_val_name' : self.npz_val_name,
                 'model_name' : self.model_name,
                 'residual' : self.residual,
                 'multievent' : self.multievent,
                 'yolo_v0': self.yolo_v0,
                 'yolo_v1': self.yolo_v1,
                 'yolo_v2': self.yolo_v2,
                 'nboxes' : self.nboxes,
                 'imagex' : self.imagex,
                 'imagey' : self.imagey,
                 'size_tminus' : self.size_tminus,
                 'size_tplus' : self.size_tplus,
                 'stride' : self.stride,
                 'depth' : self.depth,
                 'categories' : self.categories,
                 'box_vector' : self.box_vector,
                 'start_kernel' : self.start_kernel,
                 'lstm_kernel' : self.lstm_kernel,
                 'mid_kernel' : self.mid_kernel,
                 'startfilter' : self.startfilter,
                 'lstm_hidden_unit' : self.lstm_hidden_unit,
                 'epochs' : self.epochs,
                 'learning_rate' : self.learning_rate,
                 'batch_size' : self.batch_size,
                 'show' : self.show,
                 'stage_number' : self.stage_number,
                 'last_conv_factor' : self.last_conv_factor
                 }
         
         for (k,v) in self.key_categories.items():
             config[k] = v
             
         for (k,v) in self.key_cord.items():
             config[k] = v
             
             
             
         return config
         
         
        
          
    def is_valid(self, return_invalid=False):
            """Check if configuration is valid.
            Returns
            -------
            bool
            Flag that indicates whether the current configuration values are valid.
            """
            def _is_int(v,low=None,high=None):
              return (
                isinstance(v,int) and
                (True if low is None else low <= v) and
                (True if high is None else v <= high)
              )

            ok = {}
            ok['npz_directory'] = isinstance(self.npz_directory, str)
            ok['npz_name'] = isinstance(self.npz_name, str)
            ok['npz_val_name'] = isinstance(self.npz_val_name, str)
            ok['residual'] = isinstance(self.residual,bool)
            ok['yolo_v0'] = isinstance(self.yolo_v0,bool)
            ok['yolo_v1'] = isinstance(self.yolo_v1,bool)
            ok['yolo_v2'] = isinstance(self.yolo_v2,bool)
            ok['depth']         = _is_int(self.depth,1)
            ok['stride']         = _is_int(self.stride,1)
            ok['start_kernel']       = _is_int(self.start_kernel,1)
            ok['mid_kernel']         = _is_int(self.mid_kernel,1)
            ok['lstm_kernel']         = _is_int(self.lstm_kernel,1)
            ok['startfilter']        = _is_int(self.startfilter, 1)
            ok['stage_number']        = _is_int(self.stage_number, 1)
            ok['last_conv_factor']        = _is_int(self.last_conv_factor, 1)
            ok['lstm_hidden_unit']         = _is_int(self.lstm_hidden_unit,1)
            ok['epochs']        = _is_int(self.epochs, 1)
            ok['nboxes']       = _is_int(self.nboxes, 1)
            
            ok['imagex']        = _is_int(self.imagex, 1)
            ok['imagey']         = _is_int(self.imagey,1)
            ok['size_tminus']        = _is_int(self.size_tminus, 1)
            ok['size_tplus']       = _is_int(self.size_tplus, 1)
            
            ok['learning_rate'] = np.isscalar(self.learning_rate) and self.learning_rate > 0
            ok['multievent'] = isinstance(self.multievent,bool)
            ok['show'] = isinstance(self.show,bool)
            ok['categories'] =  _is_int(len(self.key_categories), 1)
            ok['box_vector'] =  _is_int(self.box_vector, 1)
            
    
            if return_invalid:
                return all(ok.values()), tuple(k for (k,v) in ok.items() if not v)
            else:
                return all(ok.values())
                   

           
           
           

        



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QComboBox, QWidget, QFormLayout, QSpinBox

from ..easing import Easing


class FrameWidget(QWidget):
    """Widget for interatviely making animations using the napari viewer."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._layout = QFormLayout(parent=self)
        self.stepsSpinBox = QSpinBox()
        self.stepsSpinBox.setValue(1)

        self.startframeSpinBox = QSpinBox()
        self.startframeSpinBox.setValue(0)

        self.endframeSpinBox = QSpinBox()
        self.endframeSpinBox.setValue(10)

        self.easeComboBox = QComboBox()
        self.easeComboBox.addItems([e.name.lower() for e in Easing])
        index = self.easeComboBox.findText('linear', Qt.MatchFixedString)
        self.easeComboBox.setCurrentIndex(index)

        self._layout.addRow('Steps', self.stepsSpinBox)
        self._layout.addRow('Ease', self.easeComboBox)
        self._layout.addRow('StartFrame', self.startframeSpinBox)
        self._layout.addRow('EndFrame', self.endframeSpinBox)



from copy import deepcopy

import numpy as np


def interpolate_state(initial_state, final_state, fraction):
    """Interpolate a state between two states

    Parameters
    ----------
    initial_state : dict
        Description of initial viewer state.
    final_state : dict
        Description of final viewer state.
    fraction : float
        Interpolation fraction, must be between `0` and `1`.
        A value of `0` will return the initial state. A
        value of `1` will return the final state.

    Returns
    -------
    state : dict
        Description of viewer state.
    """
    # Once dataclasses in napari are typed we shouldn't need to test both the
    # initial and final states. Right now we need to test both for cases
    # where one can be None.
    if isinstance(initial_state, dict) and isinstance(final_state, dict):
        state = dict()
        for k in initial_state.keys():
            v0 = initial_state[k]
            v1 = final_state[k]
            state[k] = interpolate_state(v0, v1, fraction)
        return state

    elif isinstance(initial_state, float) and isinstance(final_state, float):
        return _interpolate_float(initial_state, final_state, fraction)

    elif isinstance(initial_state, int) and isinstance(final_state, int):
        return _interpolate_int(initial_state, final_state, fraction)

    elif isinstance(initial_state, (list, tuple)) and isinstance(
        final_state, (list, tuple)
    ):
        return tuple(
            interpolate_state(v0, v1, fraction)
            for v0, v1 in zip(initial_state, final_state)
        )

    else:
        return _interpolate_bool(initial_state, final_state, fraction)


def _interpolate_float(a, b, fraction):
    return a + (b - a) * fraction


def _interpolate_int(a, b, fraction):
    return int(_interpolate_float(a, b, fraction))


def _interpolate_bool(a, b, fraction):
    if fraction < 0.5:
        return a
    else:
        return b



from copy import deepcopy
from dataclasses import asdict, is_dataclass
from pathlib import Path

from tqdm import tqdm

import imageio
import numpy as np
import skimage.io
import skimage.transform

from .utils import interpolate_state


class Animation:
    """Make animations using the napari viewer.

    Parameters
    ----------
    viewer : napari.Viewer
        napari viewer.

    Attributes
    ----------
    key_frames : list of dict
        List of viewer state dictionaries.
    frame : int
        Currently shown key frame.
    """

    def __init__(self, viewer, savedir):
        self.viewer = viewer
        self.savedir = savedir
        
        self.key_frames = []
        self.frame = -1

    def capture_keyframe(self, start, end, steps=1, ease=None, insert=True, frame=None):
        """Record current key-frame

        Parameters
        ----------
        steps : int
            Number of interpolation steps between last keyframe and captured one.
        ease : callable, optional
            If provided this method should make from `[0, 1]` to `[0, 1]` and will
            be used as an easing function for the transition between the last state
            and captured one.
        insert : bool
            If captured key-frame should insert into current list or replace the current
            keyframe.
        frame : int, optional
            If provided use this value for frame rather than current frame number.
        """
        self.key_frames = []
        for frame in tqdm(range(int(start), int(end))):
            self.frame = frame
            print(self.frame)
            new_state = {
                'viewer': self._get_viewer_state(),
                'steps': steps,
                'ease': ease,
            }
            self.key_frames.insert(self.frame + 1, new_state)
            self.viewer.dims.set_point(0, self.frame)

    def set_to_keyframe(self, frame):
        """Set the viewer to a given key-frame

        Parameters
        -------
        frame : int
            Key-frame to visualize
        """
        self.frame = frame
        if len(self.key_frames) > 0 and self.frame > -1:
            self._set_viewer_state(self.key_frames[frame])

    def _get_viewer_state(self):
        """Capture current viewer state

        Returns
        -------
        new_state : dict
            Description of viewer state.
        """
        new_state = {
            'camera': self.viewer.camera.dict(),
            'dims': self.viewer.dims.dict(),
        }
        # Log transform zoom for linear interpolation
        new_state['camera']['zoom'] = np.log10(new_state['camera']['zoom'])
        return new_state

    def _set_viewer_state(self, state):
        """Sets the current viewer state

        Parameters
        ----------
        state : dict
            Description of viewer state.
        """
        # Undo log transform zoom for linear interpolation
        camera_state = deepcopy(state['camera'])
        camera_state['zoom'] = np.power(10, camera_state['zoom'])

        self.viewer.camera.update(camera_state)
        self.viewer.dims.update(state['dims'])

    def _state_generator(self):
        if len(self.key_frames) < 2:
            raise ValueError(
                f'Must have at least 2 key frames, received {len(self.key_frames)}'
            )
        for frame in range(len(self.key_frames) - 1):
            initial_state = self.key_frames[frame]["viewer"]
            final_state = self.key_frames[frame + 1]["viewer"]
            interpolation_steps = self.key_frames[frame + 1]["steps"]
            ease = self.key_frames[frame + 1]["ease"]
            for interp in range(interpolation_steps):
                fraction = interp / interpolation_steps
                if ease is not None:
                    fraction = ease(fraction)
                state = interpolate_state(initial_state, final_state, fraction)
                yield state

    def _frame_generator(self, canvas_only=True):
        total = np.sum([f["steps"] for f in self.key_frames[1:]])
        for i, state in tqdm(enumerate(self._state_generator())):
            self._set_viewer_state(state)
            frame = self.viewer.screenshot(canvas_only=canvas_only)
            yield frame

    def animate(
        self,
        path,
        fps=20,
        quality=5,
        format=None,
        canvas_only=True,
        scale_factor=None,
    ):
        """Create a movie based on key-frames

        Parameters
        -------
        path : str
            path to use for saving the movie (can also be a path)
            should be either .mp4 or .gif. If no extension is provided,
            images are saved as a folder of PNGs
        interpolation_steps : int
            Number of steps for interpolation.
        fps : int
            frames per second
        quality: float
            number from 1 (lowest quality) to 9
            only applies to mp4
        format: str
            The format to use to write the file. By default imageio selects the appropriate for you based on the filename.
        canvas_only : bool
            If True include just includes the canvas, otherwise include the full napari viewer.
        scale_factor : float
            Rescaling factor for the image size. Only used without
            viewer (with_viewer = False).
        """

        # create a frame generator
        frame_gen = self._frame_generator(canvas_only=canvas_only)

        # create path object
        path_obj = Path(path)

        # try to create an ffmpeg writer. If not installed default to folder creation
        writer = imageio.get_writer(path, fps=fps, format=format)

        # save frames
        for ind, frame in enumerate(frame_gen):
            if scale_factor is not None:
                frame = skimage.transform.rescale(
                    frame, scale_factor, multichannel=True, preserve_range=True
                )
                frame = frame.astype(np.uint8)
            writer.append_data(frame)

        writer.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 25 13:32:04 2021

@author: vkapoor
"""
from oneat.NEATUtils import plotters
import numpy as np
from oneat.NEATUtils import helpers
from oneat.NEATUtils.helpers import load_json, yoloprediction, normalizeFloatZeroOne,microscope_dynamic_nms
from keras import callbacks
import os
import tensorflow as tf
import time
from oneat.NEATModels.nets import Concat
from tqdm import tqdm
# from IPython.display import clear_output
from pathlib import Path
from keras.models import load_model
import csv
from natsort import natsorted
import glob
import matplotlib.pyplot as plt
import h5py
import cv2
from PIL import Image
import matplotlib.pyplot as plt
from oneat.NEATModels.neat_goldstandard import NEATDynamic

class NEATPredict(NEATDynamic):
    

    def __init__(self, config, model_dir, model_name, catconfig=None, cordconfig=None):

          super().__init__(config = config, model_dir = model_dir, model_name = model_name, catconfig = catconfig, cordconfig = cordconfig)


    def predict_microscope(self, imagedir, movie_name_list, movie_input, Z_imagedir, Z_movie_name_list, Z_movie_input, start,
                Z_start, downsample=1, roi_start = 0, roi_end = 1,
                event_label_interest=1, fileextension='*TIF', nb_prediction=3, n_tiles=(1, 1), Z_n_tiles=(1, 2, 2),
                overlap_percent=0.6, event_threshold = 0.5, event_confidence = 0.5, iou_threshold=0.01, projection_model=None, delay_projection=4,
                fidelity=4, jumpindex = 1):

        self.imagedir = imagedir
        self.basedirResults = self.imagedir + '/' + "live_results"
        Path(self.basedirResults).mkdir(exist_ok=True)
        # Recurrsion variables
        self.movie_name_list = movie_name_list
        self.movie_input = movie_input
        self.Z_movie_name_list = Z_movie_name_list
        self.delay_projection = delay_projection
        self.Z_movie_input = Z_movie_input
        self.Z_imagedir = Z_imagedir
        self.start = start
        self.jumpindex = jumpindex
        self.fidelity = fidelity
        self.event_label_interest = event_label_interest
        self.Z_start = Z_start
        self.projection_model = projection_model
        self.nb_prediction = nb_prediction
        self.fileextension = fileextension
        self.n_tiles = n_tiles
        self.roi_start = roi_start
        self.roi_end = roi_end
        self.Z_n_tiles = Z_n_tiles
        self.overlap_percent = overlap_percent
        self.iou_threshold = iou_threshold
        self.event_threshold = event_threshold
        self.event_confidence = event_confidence
        self.downsample = downsample
        f = h5py.File(self.model_dir + self.model_name + '.h5', 'r+')
        data_p = f.attrs['training_config']
        data_p = data_p.decode().replace("learning_rate", "lr").encode()
        f.attrs['training_config'] = data_p
        f.close()
        self.model = load_model(self.model_dir + self.model_name + '.h5',
                                custom_objects={'loss': self.yololoss, 'Concat': Concat})

        # Z slice folder listener
        while 1:

            Z_Raw_path = os.path.join(self.Z_imagedir, self.fileextension)
            Z_filesRaw = glob.glob(Z_Raw_path)
            Z_filesRaw = natsorted(Z_filesRaw)

            Raw_path = os.path.join(self.imagedir, '*tif')
            filesRaw = glob.glob(Raw_path)
            filesRaw = natsorted(filesRaw)
            for Z_movie_name in Z_filesRaw:
                Z_Name = os.path.basename(os.path.splitext(Z_movie_name)[0])
                # Check for unique filename
                if Z_Name not in self.Z_movie_name_list and "w2" in Z_Name:
                    self.Z_movie_name_list.append(Z_Name)
                    self.Z_movie_input.append(Z_movie_name)

            for movie_name in filesRaw:
                Name = os.path.basename(os.path.splitext(movie_name)[0])
                # Check for unique filename
                if Name not in self.movie_name_list:
                    self.movie_name_list[Name] = Name
                    self.movie_input[Name] = movie_name

                    total_movies = len(self.movie_input)

            doproject = True

            for i in range(len(self.Z_movie_name_list)):

                Z_Name = self.Z_movie_name_list[i]
                Z_path = self.Z_movie_input[i]

                if Z_Name in self.movie_name_list:

                    Name = self.movie_name_list[Z_Name]
                    path = self.movie_input[Z_Name]

                    doproject = False
                else:
                    doproject = True

                if doproject:
                    time.sleep(self.delay_projection)
                    try:
                        start_time = time.time()
                        print('Reading Z stack for projection')
                        Z_image = imread(Z_path)
                        print('Read properly')
                    except:

                        Z_image = None
                    if Z_image is not None:
                        if self.projection_model is not None:
                            print('Projecting using the projection model')
                            projection = self.projection_model.predict(Z_image, 'ZYX', n_tiles=Z_n_tiles)
                        else:
                            print('Doing max projection')
                            projection = np.amax(Z_image, axis=0)
                        imwrite(self.imagedir + '/' + Z_Name + '.tif', projection.astype('float32'))
                        print(f'____ Projection took {(time.time() - start_time)} seconds ____ ')

                    else:
                        if Z_Name in self.Z_movie_name_list:
                            self.Z_movie_name_list.remove(Z_Name)
                        if Z_movie_name in self.Z_movie_input:
                            self.Z_movie_input.remove(Z_movie_name)

            self.movie_input_list = []
            for (k, v) in self.movie_input.items():
                self.movie_input_list.append(v)
            total_movies = len(self.movie_input_list)
            if total_movies > self.size_tminus + self.start:
                current_movies = imread(self.movie_input_list[self.start:self.start + self.size_tminus + 1])

                sizey = current_movies.shape[1]
                sizex = current_movies.shape[2]
                if self.downsample !=1:
                    scale_percent = 100/self.downsample
                    width = int(sizey * scale_percent / 100)
                    height = int(sizex * scale_percent / 100)
                    dim = (width, height)
                    sizex = height
                    sizey = width

                    current_movies_down = np.zeros([current_movies.shape[0], sizey, sizex])
                    # resize image
                    for j in range(current_movies.shape[0]):
                        current_movies_down[j, :] = cv2.resize(current_movies[j, :], dim, interpolation=cv2.INTER_AREA)
                else:
                    current_movies_down = current_movies
                # print(current_movies_down.shape)
                print('Predicting on Movies:', self.movie_input_list[self.start:self.start + self.size_tminus + 1])
                inputtime = self.start + self.size_tminus

                eventboxes = []
                classedboxes = {}
                smallimage = CreateVolume(current_movies_down, self.size_tminus + 1, 0)

                smallimage = normalizeFloatZeroOne(smallimage, 1, 99.8)
                # Break image into tiles if neccessary
                self.image = smallimage
                print('Doing ONEAT prediction')
                start_time = time.time()
                predictions, allx, ally = self.predict_main(smallimage)
                print(f'____ Prediction took {(time.time() - start_time)} seconds ____ ' )

                # Iterate over tiles
                for p in tqdm(range(0, len(predictions))):

                    sum_time_prediction = predictions[p]

                    if sum_time_prediction is not None:
                        for i in range(0, sum_time_prediction.shape[0]):
                            time_prediction = sum_time_prediction[i]

                            boxprediction = yoloprediction(ally[p], allx[p], time_prediction, self.stride, inputtime,
                                                           self.config, self.key_categories, self.key_cord, self.nboxes,
                                                           'prediction', 'dynamic')

                            if boxprediction is not None:
                                eventboxes = eventboxes + boxprediction

                for (event_name, event_label) in self.key_categories.items():

                    if event_label > 0:
                        current_event_box = []
                        for box in eventboxes:

                            event_prob = box[event_name]
                            event_confidence = box['confidence']
                            if event_prob >= self.event_threshold and event_confidence >= self.event_confidence:
                                current_event_box.append(box)
                        classedboxes[event_name] = [current_event_box]

                self.classedboxes = classedboxes
                self.eventboxes = eventboxes
                print('Performining non maximal supression')
                start_time = time.time()
                self.iou_classedboxes = classedboxes
                self.nms_microscope()
                print(f'____ NMS took {(time.time() - start_time)} seconds ____ ')
                print('Generating ini file')
                self.to_csv_microscope()
                self.predict(self.imagedir, self.movie_name_list, self.movie_input, self.Z_imagedir,
                             self.Z_movie_name_list, self.Z_movie_input, self.start, Z_start,
                             fileextension=self.fileextension, downsample=self.downsample,
                             nb_prediction=self.nb_prediction, n_tiles=self.n_tiles, Z_n_tiles=self.Z_n_tiles,
                             overlap_percent=self.overlap_percent, event_threshold=self.event_threshold, event_confidence = self.event_confidence,
                             iou_threshold=self.iou_threshold, projection_model=self.projection_model)

    def nms_microscope(self):
        
        
        best_iou_classedboxes = {}
        self.iou_classedboxes = {}
        self.start = self.start + self.jumpindex
        for (event_name,event_label) in self.key_categories.items():
            if event_label > 0:
               
               best_sorted_event_box = microscope_dynamic_nms( self.classedboxes, event_name, self.downsample, self.iou_threshold, self.event_threshold, self.imagex, self.imagey, self.fidelity)
               
               
               best_iou_classedboxes[event_name] = [best_sorted_event_box]
               
        self.iou_classedboxes = best_iou_classedboxes

    def to_csv_microscope(self):

        for (event_name, event_label) in self.key_categories.items():

            if event_label > 0:

                xlocations = []
                ylocations = []
                scores = []
                tlocations = []
                radiuses = []
                confidences = []

                iou_current_event_boxes = self.iou_classedboxes[event_name][0]
                bbox_left_y = self.image.shape[1] * self.roi_start
                bbox_right_y = self.image.shape[1] * self.roi_end

                bbox_left_x = self.image.shape[2] * self.roi_start
                bbox_right_x = self.image.shape[2] * self.roi_end
                iou_current_event_boxes = sorted(iou_current_event_boxes, key=lambda x: x[event_name], reverse=True)

                for iou_current_event_box in iou_current_event_boxes:
                    xcenter = iou_current_event_box['xcenter']
                    ycenter = iou_current_event_box['ycenter']
                    tcenter = iou_current_event_box['real_time_event']
                    score = iou_current_event_box[event_name]
                    radius = np.sqrt(
                        iou_current_event_box['height'] * iou_current_event_box['height'] + iou_current_event_box[
                            'width'] * iou_current_event_box['width']) // 2
                    confidence = iou_current_event_box['confidence']
                    if xcenter >= bbox_left_x and xcenter <= bbox_right_x and ycenter >= bbox_left_y and ycenter <= bbox_right_y:
                        print(round(xcenter), round(ycenter), score)
                        xlocations.append(round(xcenter))
                        ylocations.append(round(ycenter))
                        scores.append(score)
                        tlocations.append(tcenter)
                        radiuses.append(radius)
                        confidences.append(confidence)

                event_count = np.column_stack([xlocations, ylocations])
                total_event_count = np.column_stack([tlocations, ylocations, xlocations, scores, radiuses, confidences])
                csvname = self.basedirResults + "/" + event_name
                if len(xlocations) > 0:
                    writer = csv.writer(open(csvname + ".ini", 'w'))
                    writer.writerow(["[main]"])
                    writer.writerow(["nbPredictions=" + str(self.nb_prediction)])
                    live_event_data = []
                    count = 1

                    for line in event_count:
                        if len(live_event_data) > self.nb_prediction:
                            break
                        live_event_data.append(line)
                        writer.writerow(["[" + str(count) + "]"])
                        writer.writerow(["x=" + str(live_event_data[0][0])])
                        writer.writerow(["y=" + str(live_event_data[0][1])])
                        live_event_data = []

                        count = count + 1

                ImageResults = self.basedirResults + '/' + 'ImageLocations'
                Path(ImageResults).mkdir(exist_ok=True)

                event_data = []
                writer = csv.writer(open(csvname + ".csv", "a"))
                filesize = os.stat(csvname + ".csv").st_size
                if filesize < 1:
                    writer.writerow(['T', 'Y', 'X', 'Score', 'Size', 'Confidence'])
                for line in total_event_count:
                    if line not in event_data:
                        event_data.append(line)
                    writer.writerows(event_data)
                    event_data = []


def CreateVolume(patch, imaget, timepoint):
    starttime = timepoint
    endtime = timepoint + imaget
    smallimg = patch[starttime:endtime, :]

    return smallimg
